import json
import os
import time
import sys

from watchdog.observers import Observer
from watchdog.tricks import Trick
import cloudbot


class Config(dict):
    """
    :type filename: str
    :type path: str
    :type bot: cloudbot.core.bot.CloudBot
    :type observer: Observer
    :type event_handler: ConfigEventHandler
    """

    def __init__(self, bot, *args, **kwargs):
        """
        :type bot: cloudbot.core.bot.CloudBot
        :type args: list
        :type kwargs: dict
        """
        super().__init__(*args, **kwargs)
        self.filename = "config.json"
        self.path = os.path.abspath(self.filename)
        self.bot = bot
        self.logger = bot.logger
        self.update(*args, **kwargs)

        # populate self with config data
        self.load_config()

        if cloudbot.dev_mode["config_reloading"]:
            # start watcher
            self.observer = Observer()

            pattern = "*{}".format(self.filename)

            self.event_handler = ConfigEventHandler(self.bot, self, patterns=[pattern])
            self.observer.schedule(self.event_handler, path='.', recursive=False)
            self.observer.start()

    def stop(self):
        """shuts down the config reloader"""
        if cloudbot.dev_mode["config_reloading"]:
            self.observer.stop()

    def load_config(self):
        """(re)loads the bot config from the config file"""
        if not os.path.exists(self.path):
            # if there is no config, show an error and die
            self.logger.critical("No config file found, bot shutting down!")
            print("No config file found! Bot shutting down in five seconds.")
            print("Copy 'config.default' to 'config.json' for defaults.")
            print("For help, see http://git.io/cloudbotirc. Thank you for using CloudBot!")
            time.sleep(5)
            sys.exit()

        with open(self.path) as f:
            self.update(json.load(f))
            self.logger.debug("Config loaded from file.")

        # reload permissions
        if self.bot.connections:
            for connection in self.bot.connections:
                connection.permissions.reload()

    def save_config(self):
        """saves the contents of the config dict to the config file"""
        json.dump(self, open(self.path, 'w'), sort_keys=True, indent=4)
        self.logger.info("Config saved to file.")


class ConfigEventHandler(Trick):
    """
    :type bot: cloudbot.core.bot.CloudBot
    :type config: core.config.Config
    :type logger: logging.Logger
    """

    def __init__(self, bot, config, *args, **kwargs):
        """
        :type bot: cloudbot.core.bot.CloudBot
        :type config: Config
        """
        self.bot = bot
        self.config = config
        self.logger = config.logger
        Trick.__init__(self, *args, **kwargs)

    def on_any_event(self, event):
        if self.bot.running:
            self.logger.info("Config changed, triggering reload.")
            self.config.load_config()
